@@ -5,7 +5,7 @@
  * Written by Stephane Bortzmeyer <bortz@users.sourceforge.net>. See
  * the AUTHORS file for other contributors. 
  * 
- * $Id: echoping.c,v 4.25 2002/11/04 15:21:21 bortz Exp $
+ * $Id: echoping.c,v 4.37 2004/02/20 13:05:27 bortz Exp $
  *
  *  */
 
@@ -18,7 +18,7 @@ unsigned short timeout_flag;
 /*
  * An option to define only if you want to drive echoping from another
  * process. Useless but harmless otherwise. In practice, while OSF/1 is happy
- * with it, SunOS refuses to use fflush on a NULL and Linux fails.
+ * with it, SunOS refuses to use fflush on a NULL and Linux fails. 
  */
 #undef FLUSH_OUTPUT
 
@@ -50,7 +50,7 @@ main (argc, argv)
   int result;
 
   int sockfd;
-  struct addrinfo hints, *res;
+  struct addrinfo hints, hints_numeric, *res;
   int error;
   char hbuf[NI_MAXHOST], pbuf[NI_MAXSERV];
 #ifdef NI_WITHSCOPEID
@@ -119,6 +119,14 @@ main (argc, argv)
   SSL_CTX *ctx = NULL;
   SSL *sslh = NULL;
 #endif
+#ifdef GNUTLS
+  gnutls_session session;
+  gnutls_certificate_credentials xcred;
+  int tls_result;
+  const int cert_type_priority[3] = { GNUTLS_CRT_X509,
+    GNUTLS_CRT_OPENPGP, 0
+  };
+#endif
 
   int priority;
   int priority_requested = 0;
@@ -328,11 +336,11 @@ main (argc, argv)
       exit (1);
     }
 #endif
-#ifndef OPENSSL
+#if ! (defined(OPENSSL) || defined(GNUTLS))
   if (ssl)
     {
       (void) fprintf (stderr,
-		      "%s: not compiled with SSL support.\n", progname);
+		      "%s: not compiled with SSL/TLS support.\n", progname);
       exit (1);
     }
 #endif
@@ -417,6 +425,15 @@ main (argc, argv)
       printf ("\nThis is %s, version %s.\n\n", progname, VERSION);
     }
   server = argv[0];
+#ifdef LIBIDN
+  locale_server = server;
+  utf8_server = stringprep_locale_to_utf8 (server);
+  if (utf8_server)
+    server = utf8_server;
+  else
+    err_quit ("Cannot convert %s to UTF-8 encoding: wrong locale (%s)?",
+	      server, stringprep_locale_charset ());
+#endif
   if (!http && !icp)
     {
       for (p = server; *p && (*p != ':'); p++)
@@ -508,6 +525,33 @@ main (argc, argv)
     }
 #endif
 
+#ifdef LIBIDN
+  /* Check if it is an address or a name (libidn will have trouble with 
+     IPv6 addresses otherwise) */
+  memset (&hints_numeric, 0, sizeof (hints_numeric));
+  hints_numeric.ai_family = family;
+  hints_numeric.ai_flags = AI_NUMERICHOST;
+  error = getaddrinfo (server, port_name, &hints_numeric, &res);
+  if (error && error == EAI_NONAME)	/* A name, not an address */
+    {
+      if ((result =
+	   idna_to_ascii_8z (utf8_server, &ace_server,
+			     IDNA_USE_STD3_ASCII_RULES)) != IDNA_SUCCESS)
+	{
+	  if (result == IDNA_CONTAINS_LDH) 
+	    err_quit ("Illegal name for host: %s", server); /* foo@bar or 
+                                                               similar errors */
+	  else
+	    err_quit ("IDN error for host: %s %d", server, result);
+	}
+      if (strcmp (utf8_server, ace_server))
+	{
+	  if (verbose)
+	    printf ("ACE name of the server: %s\n", ace_server);
+	  server = ace_server;
+	}
+    }				/* Else it is an address, do not IDNize it */
+#endif
   error = getaddrinfo (server, port_name, &hints, &res);
   if (error)
     {
@@ -530,7 +574,8 @@ main (argc, argv)
 #ifdef HTTP
   if (http)
     {
-      sendline = make_http_sendline (url, server, res->ai_protocol, nocache);
+      sendline = make_http_sendline (url, server, atoi (pbuf), nocache);
+      /* printf ("DEBUG: sending %s\n", sendline); */
     }
   else
 #endif
@@ -591,11 +636,19 @@ main (argc, argv)
       RAND_file_name (rand_file, sizeof (rand_file));
       RAND_write_file (rand_file);
       RAND_load_file (rand_file, 1024);
-      meth = SSLv2_client_method ();
+      meth = SSLv23_client_method ();
       if ((ctx = SSL_CTX_new (meth)) == NULL)
 	err_sys ("Cannot create a new SSL context");
     }
 #endif
+#ifdef GNUTLS
+  if (ssl)
+    {
+      gnutls_global_init ();
+      gnutls_certificate_allocate_credentials (&xcred);
+      /* gnutls_certificate_set_x509_trust_file(xcred, CAFILE, GNUTLS_X509_FMT_PEM); */
+    }
+#endif
 
   for (i = 1; i <= number; i++)
     {
@@ -791,6 +844,40 @@ main (argc, argv)
 	      /* We could check the server's certificate or other funny
 	         things */
 	    }
+#endif
+#ifdef GNUTLS
+	  if (ssl)
+	    {
+	      tls_result = gnutls_init (&session, GNUTLS_CLIENT);
+	      if (tls_result != 0)
+		err_sys ("Cannot create a new TLS session");
+	      gnutls_set_default_priority (session);
+	      gnutls_certificate_type_set_priority (session,
+						    cert_type_priority);
+	      gnutls_credentials_set (session, GNUTLS_CRD_CERTIFICATE, xcred);
+	      gnutls_transport_set_ptr (session,
+					(gnutls_transport_ptr) sockfd);
+	      tls_result = gnutls_handshake (session);
+	      if (tls_result < 0)
+		{
+		  if ((errno == EINTR) && (timeout_flag))
+		    {
+		      printf ("Timeout while starting TLS\n");
+		      close (sockfd);
+		      continue;
+		    }
+		  else
+		    {
+		      err_sys ("Cannot start the TLS session: %s",
+			       gnutls_strerror (tls_result));
+		    }
+		}
+	      if (verbose)
+		printf ("TLS connection using \"%s\"\n",
+			gnutls_cipher_get_name (gnutls_cipher_get (session)));
+	      /* We could check the server's certificate or other funny
+	         things. See http://www.gnu.org/software/gnutls/documentation/gnutls/gnutls.html#SECTION00622000000000000000 */
+	    }
 #endif
 	}
       /* Not T/TCP */
@@ -854,6 +941,29 @@ main (argc, argv)
 		    }
 		  /* printf ("DEBUG: writing %s with SSL\n", sendline); */
 		}
+#endif
+#ifdef GNUTLS
+	      else
+	      {
+		if ((rc =
+		     gnutls_record_send (session, sendline,
+					 strlen (sendline))) != n)
+		  {
+		    if ((nr < 0 || nr != n) && timeout_flag)
+		      {
+			nr = n;
+			printf ("Timeout while writing\n");
+			close (sockfd);
+			continue;
+		      }
+		    else
+		      {
+			err_sys ("gnutls_record_send error %d on socket: %s",
+				 rc, gnutls_strerror (rc));
+		      }
+		  }
+		/* printf ("DEBUG: writing %s with TLS\n", sendline); */
+	      }
 #endif
 	      /* Write something to the server */
 	      if (writen (sockfd, sendline, n) != n)
@@ -922,7 +1032,12 @@ main (argc, argv)
 	      n = SSL_get_fd (sslh);
 	    }
 #endif
-
+#ifdef GNUTLS
+	  else
+	  {
+	    n = sockfd;
+	  }
+#endif
 	  FD_SET (n, &mask);
 	  if (select (n + 1, &mask, 0, 0, NULL) > 0)
 	    {
@@ -959,6 +1074,10 @@ main (argc, argv)
 #ifdef OPENSSL
 		  else
 		    channel.ssl = sslh;
+#endif
+#ifdef GNUTLS
+		  else
+		  channel.tls = session;
 #endif
 		  nr = read_from_server (channel, ssl);
 		}
@@ -970,7 +1089,6 @@ main (argc, argv)
 		}
 #endif
 
-
 	    }
 	  else
 	    {
@@ -1082,6 +1200,11 @@ main (argc, argv)
 	  if (ssl)
 	    SSL_shutdown (channel.ssl);
 	  else
+#endif
+#ifdef GNUTLS
+	  if (ssl)
+	    shutdown (sockfd, SHUT_RDWR);
+	  else
 #endif
 	    fclose (channel.fs);
 	}
@@ -1154,15 +1277,29 @@ main (argc, argv)
 	      /* SSL_clear (sslh); No, we have to free. Bug #130151 */
 	      SSL_free (sslh);
 	    }
+#endif
+#ifdef GNUTLS
+	  if (ssl)
+	    {
+	      gnutls_bye (channel.tls, GNUTLS_SHUT_RDWR);
+	      gnutls_deinit (session);
+	      /* gnutls_certificate_free_credentials(xcred); */
+	    }
 #endif
 	}
-    }
+    }				/* End of main loop */
   printstats ();
   if (successes >= 1)
     exit (0);
   else
     exit (1);
-  /* It would be nice to clean here (SSL, etc) */
+  /* It would be nice to clean here (OpenSSL, etc) */
+#ifdef GNUTLS
+  if (ssl)
+    {
+      gnutls_global_deinit ();
+    }
+#endif
 }
 
 
